from quandl.errors.quandl_error import InvalidRequestError
from .model.dataset import Dataset
from .model.merged_dataset import MergedDataset
from .utils.api_key_util import ApiKeyUtil
from .message import Message
from six import string_types
import warnings

OLD_TO_NEW_PARAMS = {'authtoken': 'api_key', 'trim_start': 'start_date',
                     'trim_end': 'end_date', 'transformation': 'transform',
                     'sort_order': 'order'}


def get(dataset, **kwargs):
    """Return dataframe of requested dataset from Quandl.
    :param dataset: str or list, depending on single dataset usage or multiset usage
            Dataset codes are available on the Quandl website
    :param str api_key: Downloads are limited to 50 unless api_key is specified
    :param str start_date, end_date: Optional datefilers, otherwise entire
           dataset is returned
    :param str collapse: Options are daily, weekly, monthly, quarterly, annual
    :param str transform: options are diff, rdiff, cumul, and normalize
    :param int rows: Number of rows which will be returned
    :param str order: options are asc, desc. Default: `asc`
    :param str returns: specify what format you wish your dataset returned as,
        either `numpy` for a numpy ndarray or `pandas`. Default: `pandas`
    :returns: :class:`pandas.DataFrame` or :class:`numpy.ndarray`
    Note that Pandas expects timeseries data to be sorted ascending for most
    timeseries functionality to work.
    Any other `kwargs` passed to `get` are sent as field/value params to Quandl
    with no interference.
    """

    _convert_params_to_v3(kwargs)

    data_format = kwargs.pop('returns', 'pandas')

    ApiKeyUtil.init_api_key_from_args(kwargs)

    # Check whether dataset is given as a string
    # (for a single dataset) or an array (for a multiset call)

    # Unicode String
    if isinstance(dataset, string_types):
        dataset_args = _parse_dataset_code(dataset)
        if dataset_args['column_index'] is not None:
            kwargs.update({'column_index': dataset_args['column_index']})
        data = Dataset(dataset_args['code']).data(params=kwargs, handle_column_not_found=True)
    # Array
    elif isinstance(dataset, list):
        args = _build_merged_dataset_args(dataset)
        # handle_not_found_error if set to True will add an empty DataFrame
        # for a non-existent dataset instead of raising an error
        data = MergedDataset(args).data(params=kwargs,
                                        handle_not_found_error=True,
                                        handle_column_not_found=True)
    # If wrong format
    else:
        raise InvalidRequestError(Message.ERROR_DATASET_FORMAT)

    if data_format == 'numpy':
        return data.to_numpy()
    return data.to_pandas()


def _parse_dataset_code(dataset):
    if '.' not in dataset:
        return {'code': dataset, 'column_index': None}
    dataset_temp = dataset.split('.')
    if not dataset_temp[1].isdigit():
        raise ValueError(Message.ERROR_COLUMN_INDEX_TYPE % dataset)
    return {'code': dataset_temp[0], 'column_index': int(dataset_temp[1])}


def _build_merged_dataset_args(datasets):
    merged_dataset_args = []
    for dataset in datasets:
        dataset_code_column = _parse_dataset_code(dataset)
        arg = dataset_code_column['code']
        column_index = dataset_code_column['column_index']
        if column_index is not None:
            arg = (dataset_code_column['code'], {'column_index': [column_index]})
        merged_dataset_args.append(arg)
    return merged_dataset_args


def _convert_params_to_v3(params):
    for k, v in OLD_TO_NEW_PARAMS.items():
        if k in params:
            msg = Message.WARN_PARAMS_NOT_SUPPORTED % (k, v)
            warnings.warn(msg, DeprecationWarning)
            # update to the new query param if not specified already
            if v not in params:
                params[v] = params.pop(k)



import unittest
from test.helpers.httpretty_extension import httpretty
from test.helpers.merged_datasets_helper import setupDatasetsTest
import pandas
import numpy
from mock import patch, call, Mock
from quandl.model.dataset import Dataset
from quandl.model.merged_dataset import MergedDataset
from quandl.get import get
from quandl.api_config import ApiConfig
from quandl.connection import Connection


class GetSingleDatasetTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        # ensure api key is not set
        ApiConfig.api_key = None
        setupDatasetsTest(self, httpretty)

    @classmethod
    def tearDownClass(cls):
        httpretty.disable()
        httpretty.reset()

    def test_returns_pandas_by_default(self):
        result = get('NSE/OIL')
        self.assertIsInstance(result, pandas.core.frame.DataFrame)

    def test_returns_pandas_when_requested(self):
        result = get('NSE/OIL', returns='pandas')
        self.assertIsInstance(result, pandas.core.frame.DataFrame)

    def test_returns_numpys_when_requested(self):
        result = get('NSE/OIL', returns='numpy')
        self.assertIsInstance(result, numpy.core.records.recarray)

    def test_setting_api_key_config(self):
        mock_connection = Mock(wraps=Connection)
        with patch('quandl.connection.Connection.execute_request',
                   new=mock_connection.execute_request) as mock:
            ApiConfig.api_key = 'api_key_configured'
            get('NSE/OIL')
            # extract the headers passed to execute_request
            actual_request_headers = mock.call_args[1]['headers']
            self.assertEqual(actual_request_headers['x-api-token'], 'api_key_configured')

    def test_sets_api_key_using_authtoken_arg(self):
        get('NSE/OIL', authtoken='api_key')
        self.assertEqual(ApiConfig.api_key, 'api_key')

    def test_sets_api_key_using_api_key_arg(self):
        get('NSE/OIL', api_key='api_key')
        self.assertEqual(ApiConfig.api_key, 'api_key')

    @patch.object(Dataset, 'data')
    def test_query_params_are_formed_with_old_arg_names(self, mock_method):
        get('NSE/OIL', authtoken='authtoken', trim_start='2001-01-01',
            trim_end='2010-01-01', collapse='annual',
            transformation='rdiff', rows=4, sort_order='desc')
        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(mock_method.mock_calls[0],
                         call(handle_column_not_found=True,
                              params={'start_date': '2001-01-01', 'end_date': '2010-01-01',
                                      'collapse': 'annual', 'transform': 'rdiff',
                                      'rows': 4, 'order': 'desc'}))

    @patch.object(Dataset, 'data')
    def test_query_params_are_formed_with_new_arg_names(self, mock_method):
        get('NSE/OIL', api_key='authtoken', start_date='2001-01-01',
            end_date='2010-01-01', collapse='annual',
            transform='rdiff', rows=4, order='desc')
        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(mock_method.mock_calls[0],
                         call(handle_column_not_found=True,
                              params={'start_date': '2001-01-01', 'end_date': '2010-01-01',
                                      'collapse': 'annual', 'transform': 'rdiff',
                                      'rows': 4, 'order': 'desc'}))

    @patch('quandl.model.data.Data.all')
    def test_code_is_parsed(self, mock):
        get('NSE/OIL')
        expected = call(
            params={'dataset_code': 'OIL', 'order': 'asc',
                    'database_code': 'NSE'})
        self.assertEqual(mock.call_args, expected)

    @patch.object(Dataset, 'data')
    def test_number_becomes_column_index(self, mock_method):
        get('NSE/OIL.1')
        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(mock_method.mock_calls[0],
                         call(handle_column_not_found=True, params={'column_index': 1}))

    @patch('quandl.model.data.Data.all')
    def test_code_and_column_is_parsed_and_used(self, mock):
        get('NSE/OIL.1')
        expected = call(
            params={'dataset_code': 'OIL', 'order': 'asc',
                    'database_code': 'NSE', 'column_index': 1})
        self.assertEqual(mock.call_args, expected)

    def test_raise_error_when_non_numeric_column_index(self):
        self.assertRaises(ValueError, lambda: get('NSE/OIL.notanumber'))


class GetMultipleDatasetsTest(unittest.TestCase):
    @classmethod
    def setUp(self):
        # ensure api key is not set
        ApiConfig.api_key = None
        setupDatasetsTest(self, httpretty)

    @classmethod
    def tearDownClass(cls):
        httpretty.disable()
        httpretty.reset()

    @patch('quandl.model.merged_dataset.MergedDataset._build_dataset_object')
    def test_multiple_datasets_args_formed(self, mock):
        # requested_column_indexes is a dynamically added attribute
        self.oil_obj.requested_column_indexes = []
        mock.return_value = self.oil_obj
        get(['WIKI/AAPL.1', 'WIKI/MSFT.2', 'NSE/OIL'])
        expected = [call(('WIKI/AAPL', {'column_index': [1]})),
                    call(('WIKI/MSFT', {'column_index': [2]})),
                    call('NSE/OIL')]
        self.assertEqual(mock.call_args_list, expected)

    @patch.object(MergedDataset, 'data')
    def test_query_params_are_formed_with_old_arg_names(self, mock_method):
        get(['WIKI/AAPL.1', 'WIKI/MSFT.2', 'NSE/OIL'],
            authtoken='authtoken', trim_start='2001-01-01',
            trim_end='2010-01-01', collapse='annual',
            transformation='rdiff', rows=4, sort_order='desc')

        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(mock_method.mock_calls[0],
                         call(handle_not_found_error=True, handle_column_not_found=True,
                              params={'start_date': '2001-01-01', 'end_date': '2010-01-01',
                                      'collapse': 'annual', 'transform': 'rdiff',
                                      'rows': 4, 'order': 'desc'}))

    @patch.object(MergedDataset, 'data')
    def test_query_params_are_formed_with_new_arg_names(self, mock_method):
        get(['WIKI/AAPL.1', 'WIKI/MSFT.2', 'NSE/OIL'],
            api_key='authtoken', start_date='2001-01-01',
            end_date='2010-01-01', collapse='annual',
            transform='rdiff', rows=4, order='desc')
        self.assertEqual(mock_method.call_count, 1)
        self.assertEqual(mock_method.mock_calls[0],
                         call(handle_not_found_error=True, handle_column_not_found=True,
                              params={'start_date': '2001-01-01', 'end_date': '2010-01-01',
                                      'collapse': 'annual', 'transform': 'rdiff',
                                      'rows': 4, 'order': 'desc'}))



try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

from quandl.api_config import ApiConfig


class RequestType(object):
    """ Determines whether a request should be made using a GET or a POST request.
    Default limit of 8000 is set here as it appears to be the maximum for many
    webservers.
    """
    MAX_URL_LENGTH_FOR_GET = 8000
    USE_GET_REQUEST = True  # This is used to simplify testing code

    @classmethod
    def get_request_type(cls, url, **params):
        query_string = urlencode(params['params'])
        request_url = '%s/%s/%s' % (ApiConfig.api_base, url, query_string)
        if RequestType.USE_GET_REQUEST and (len(request_url) < cls.MAX_URL_LENGTH_FOR_GET):
            return 'get'
        else:
            return 'post'
